# Generated by Django 4.2.7 on 2023-11-19 22:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0018_searchmodelconfig_delete_searchmodel"),
    ]

    operations = [
        migrations.AlterField(
            model_name="googleuser",
            name="family_name",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
        migrations.AlterField(
            model_name="googleuser",
            name="given_name",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
        migrations.AlterField(
            model_name="googleuser",
            name="name",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
    ]
